class MultiMatrixException < Exception
end

# An n-dimensonal matrix, where dimensions are labels
#
# Imagine: your urgent need is to store a dataset more than 1 or 2 dimensions large.
# The default way to do this is tu use multi dimensional arrays, like "array[][][]" for
# example. Inserting or retrieveing a value from this dataset would be kind of an easy task,
# but what happens when You have to summarize all the elements of a dimension. For example
# sum all the elements where the second level array's index is 2. You have to loop through
# n*n*n elements, just to find some of them. Tedious.
#
# MultiMatrix comes to the rescue.
#
# MultiMatrix makes it easy to create multi-level data structures, and makes finding and looping through
# them nice and easy. And fast.
#
# @author Gergely Dömsödi <mailto:gergely.domsodi@uhusystems.com>
#
# To easily show an example, and concept of this class let's see a 2 dimensional array
#
# m = MultiMatrix.new(:x, :y)
# m.insert(:x => 1, :y = 3) = 10
# m.insert(:x => 2, :y = 1) = 20
# m.insert(:x => 2, :y =>2) = 30
#
# could be imagined as
#    x| 1 | 2 | 3 |
#   y |___|___|___|
#   1 | ? | 20| ? |
#     |___|___|___|
#   2 | ? | 30| ? |
#     |___|___|___|
#   3 | 10| ? | ? |
#     |___|___|___|
#
#  >> m.sum(x: 2)
#  => 50
#  >> m.sum(x: 3)
#  => 0
#  >> m.to_a(y: 2)
#  => [30]
#  >> m[:x => 1, :y => 1]
#  => nil
#  >> m[:x => 3] == MultiMatrix.new(:y)
#  => true
#
# Usage:
#
#  >> x = MultiMatrix.new(:x, :y, :z)
#  >> x[:x => 1, :y => 1, :z => 1] = 1
#  >> x[:x => 1, :y => 1, :z => 2] = 1
#  >> x[:x => 2, :y => 1, :z => 2] = 1
#
#  >> x[:x => 1, :y => 1, :z => 1]
#  => 1
#
#  >> x[:x => 1, :y => 1]
#  => MultiMatrix({:z=>2}=>2)
#
#  >> x.sum(:x => 1)
#  => 2
#  >> x.sum(:z => 1)
#  => 1
#  >> x.to_a(:z => 1)
#  => [1]
#
# Wait! There is more:
#
#  >> x[:z => [1,2]]
#  => [1,1,1]
#  >> x.sum(:z => [1,2])
#  => 3

class MultiMatrix

  # Creates a new MultiMatrix object, where labels will be the dimensions of the dataset
  # use labels appropriate for Hash keys, because they will be stored as such internally.
  # labels could be Symbols, Integers, even ActiveModel objects, the only requirement for it that it have to
  # have a meaningful == method.
  # @param labelskeys an [Array] of values to be used as dimension labels
  def initialize(*labelskeys)
    raise MultiMatrixException, "no dimensions" if labelskeys.length == 0
    @labels = labelskeys
    @dim = @labels.length
    @values = Hash.new
    @values_by_id = Hash.new
    @labels_by_id = Hash.new
    @sumcache = {}
    @idscache = {}
    @seq = 0
    @finderhash = {}
  end

  # Inserts a value to the data set.
  #
  # @param labels [Hash] the position where to insert the value.
  #   beware, at insert all dimension of the positin must be given
  # @param value The value to be inserted
  # @return [true] on success, raises MultiMatrixException otherwise
  def insert(labels, value)
    if check_labels(labels) and labels.keys.length == @dim
      if @values.key?(labels)
        id = @values[labels]
        if block_given?
          save_value(labels, yield(get_value(labels), value), :override => id)
        else
          save_value(labels, get_value(labels) + value, :override => id)
        end
      else
        save_value(labels, value)
      end
    else
      raise MultiMatrixException, 'not enough key supplied'
    end
    @sumcache = {}
    true
  end

  # Executes block on all value at position given by labels
  # @param labels finder hash
  def each(labels = {})
    if check_labels(labels) and labels.keys.length <= @dim
      ids = find_ids(labels)
      ids.each do |id|
        if block_given?
          yield @values_by_id[id]
        end
      end
    end
  end

  # Same as Enumerable#inject, but filter hash is accepted
  # @param labels finder hash
  # @param default beginning value of the memo
  def inject(labels = {}, default = 0)
    memo = default
    if check_labels(labels) and labels.keys.length <= @dim
      ids = find_ids(labels)
      ids.each do |id|
        if block_given?
          memo = yield memo, @values_by_id[id]
        end
      end
    end
    return memo
  end

  # Returns an array of values matching @param labels
  # @param labels finder hash
  # @return [Array] values found
  def to_a(labels = {})
    if check_labels(labels)
      ids=find_ids(labels)
      return ids.map{|id| @values_by_id[id]}
    end
  end

  # Summarizes all elements matching labels
  # a little faster than using MultiMatrix#inject(labels) {|a,b| a+b} because of two factors:
  # - if every dimension is given, sum return only one element without looping through the data structure
  # - sum is cached so if sum is called twice with the same selector hash, the result is given back instantly
  # @param labels finder hash
  # @param default default value of sum
  def sum(labels = {}, default = 0)
    if check_labels(labels)
      if labels.keys.length == @dim and labels.values.find_all {|val| val.class.to_s == 'Array'}.empty?
        return get_value(labels)
      end
      if @sumcache.key?(labels)
        return @sumcache[labels]
      end
      vals = find_ids(labels).map {|id| @values_by_id[id]}
      ret = default
      vals.each do |v| ret +=v end
      @sumcache[labels] = ret
      ret
    end
  end

  # Retrieves a value
  # If every dimension is given, retrieve just a value at the specified position,
  # if not, retrieves an MultiMatrix of the matching elements, leaving the fixed dimensions behind
  # @param labels finder hash
  # @return value/MultiMatrix
  def retrieve(labels = {})
    if check_labels(labels)
      if labels.keys.length < @dim or labels.values.find_all {|val| val.class == Array}.length > 0
        tmp = MultiMatrix.new(*(@labels - labels.keys.find_all {|val| labels[val].class != Array}))
        ids = find_ids(labels)
        ids.each do |id|
          hash = @labels_by_id[id].clone
          labels.keys.find_all {|val| labels[val].class != Array}.each do |key|
            hash.delete(key)
          end
          tmp[hash] = @values_by_id[id]
        end
        return tmp
      elsif labels.keys.length == @dim
        return get_value(labels)
      end
    end
  end

  # return the possible values of a dimension
  # where values match the labels given in [rest]
  # @param needle a key of the labels hash
  # @param rest a selector hash
  def values_for_label(needle, rest = {})
    raise MultiMatrixException, 'labels don\'t match: '+rest.keys.join(",") unless check_labels(rest)
    raise MultiMatrixException, 'label does not exist' unless check_labels({needle => nil})
    ids = find_ids(rest)
    ids.map{|id| @labels_by_id[id][needle]}
  end

  alias_method :[], :retrieve
  alias_method :[]=, :insert

  # Test the equality of two NMatrices

  # NMatrices are equal if they have the same labels,
  # label values, and  all their respective values are equal
  # @param other [MultiMatrix]
  def ==(other)
    [[self, other],[other,self]].each do |x|
      x.first.keys do |k|
        if !x.last.key?(k)
          return false
        end
        if x.last[k] != x.first[k]
          return false
        end
      end
    end
    return true
  end

  # @return labels in MultiMatrix
  def keys
    @values.keys
  end

  # @return  String representation of an MultiMatrix - shown as a Hash of label => value pairs
  def to_s
    hash = {}
    @values.each_pair do |k,v|
      hash[k] = @values_by_id[v]
    end
    hash.to_s
  end

  private

  def check_labels(to_check)
    raise MultiMatrixException, 'labels not a hash' if to_check.class != Hash
    raise MultiMatrixException, 'labels hash contains an unknown label' unless (to_check.keys-@labels).empty?
    return true;
  end

  def get_value(labels)
    @values_by_id[@values[labels]]
  end

  def save_value(labels, val, options = {})
    if options.key?(:override)
      @values_by_id[options[:override]] = val
    else
      id = generate_id
      @values[labels] = id
      @values_by_id[id] = val
      @labels_by_id[id] = labels
      labels.each_pair do |k, v|
        @finderhash[k] ||= {}
        @finderhash[k][v] ||= []
        @finderhash[k][v].push(id)
      end
    end
  end

  def find_ids(labels)
    return @idscache[labels] if !@idscache[labels].nil?
    ids = @values.values
    labels.each_pair do |k,v|
      if v.class == Array
        ids &= v.map {|v| @finderhash[k][v]}.flatten.uniq
      elsif @finderhash[k][v].nil?
        return []
      else
        ids &= @finderhash[k][v]
      end
    end
    @idscache[labels] = ids
    return ids
  end

  def generate_id
    @seq = @seq.succ
    return @seq
  end

end
